#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Flowsheet example full Water Resource Recovery Facility
(WRRF; a.k.a., wastewater treatment plant) with ASM1 and ADM1.

The flowsheet follows the same formulation as benchmark simulation model no.2 (BSM2)
but comprises different specifications for default values than BSM2.

"""
__author__ = "Alejandro Garciadiego, Adam Atia, Marcus Holly, Chenyu Wang, Ben Knueven, Xinhong Liu,"

import pyomo.environ as pyo

from pyomo.network import Arc, SequentialDecomposition
from watertap.unit_models.anaerobic_digester import AD, ADScaler
from watertap.unit_models.thickener import Thickener, ThickenerScaler
from watertap.unit_models.dewatering import DewateringUnit, DewatererScaler
from watertap.unit_models.cstr import CSTR, CSTRScaler
from watertap.unit_models.clarifier import Clarifier, ClarifierScaler

from watertap.unit_models.translators.translator_asm1_adm1 import (
    Translator_ASM1_ADM1,
    ASM1ADM1Scaler,
)
from watertap.unit_models.translators.translator_adm1_asm1 import (
    Translator_ADM1_ASM1,
    ADM1ASM1Scaler,
)

from watertap.core.solvers import get_solver
from idaes.core.initialization import BlockTriangularizationInitializer
from idaes.core.scaling import set_scaling_factor
from idaes.core.scaling.custom_scaler_base import (
    CustomScalerBase,
    ConstraintScalingScheme,
)
from idaes.core.util.exceptions import InitializationError
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties import (
    ADM1ParameterBlock,
)
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_reactions import (
    ADM1ReactionParameterBlock,
)
from idaes.models.unit_models.mixer import MomentumMixingType
from idaes.models.unit_models.separator import SplittingType
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties_vapor import (
    ADM1_vaporParameterBlock,
)

from idaes.core import FlowsheetBlock, UnitModelCostingBlock, UnitModelBlockData
from idaes.models.unit_models import (
    Feed,
    Mixer,
    Separator,
    PressureChanger,
    Product,
)
import idaes.logger as idaeslog

from watertap.unit_models.aeration_tank import (
    AerationTank,
    AerationTankScaler,
    ElectricityConsumption,
)
from watertap.property_models.unit_specific.activated_sludge.asm1_properties import (
    ASM1ParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.asm1_reactions import (
    ASM1ReactionParameterBlock,
)
from watertap.costing import WaterTAPCosting
from watertap.costing.unit_models.clarifier import (
    cost_circular_clarifier,
    cost_primary_clarifier,
)

# Set up logger
_log = idaeslog.getLogger(__name__)


def main(reactor_volume_equalities=True):
    m = build()
    set_operating_conditions(m)
    initialize_system(m)
    add_costing(m)
    m.fs.costing.initialize()

    scale_system(m)
    scaling = pyo.TransformationFactory("core.scale_model")
    scaled_model = scaling.create_using(m, rename=False)
    solve(scaled_model)
    scaling.propagate_solution(scaled_model, m)

    print("\n\n=============SIMULATION RESULTS=============\n\n")
    # display_results(m)
    display_costing(m)

    setup_optimization(m, reactor_volume_equalities=reactor_volume_equalities)
    rescale_system(m)
    rescaling = pyo.TransformationFactory("core.scale_model")
    rescaled_model = rescaling.create_using(m, rename=False)
    solve(rescaled_model, tee=True)

    results = rescaling.propagate_solution(rescaled_model, m)

    print("\n\n=============OPTIMIZATION RESULTS=============\n\n")
    # display_results(m)
    display_costing(m)
    display_performance_metrics(m)

    return m, results, rescaled_model


def build():
    m = pyo.ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props_ASM1 = ASM1ParameterBlock()
    m.fs.props_ADM1 = ADM1ParameterBlock()
    m.fs.props_vap = ADM1_vaporParameterBlock()
    m.fs.ADM1_rxn_props = ADM1ReactionParameterBlock(property_package=m.fs.props_ADM1)
    m.fs.ASM1_rxn_props = ASM1ReactionParameterBlock(property_package=m.fs.props_ASM1)
    # Feed water stream
    m.fs.FeedWater = Feed(property_package=m.fs.props_ASM1)

    # ==========================================================================
    # Activated Sludge Process
    # ==========================================================================
    # Mixer for inlet water and recycled sludge
    m.fs.MX1 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["feed_water", "recycle"],
        momentum_mixing_type=MomentumMixingType.none,
    )
    # First reactor (anoxic) - standard CSTR
    m.fs.R1 = CSTR(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )
    # Second reactor (anoxic) - standard CSTR
    m.fs.R2 = CSTR(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )
    # Third reactor (aerobic) - CSTR with injection
    m.fs.R3 = AerationTank(
        property_package=m.fs.props_ASM1,
        reaction_package=m.fs.ASM1_rxn_props,
        electricity_consumption=ElectricityConsumption.calculated,
    )
    # Fourth reactor (aerobic) - CSTR with injection
    m.fs.R4 = AerationTank(
        property_package=m.fs.props_ASM1,
        reaction_package=m.fs.ASM1_rxn_props,
        electricity_consumption=ElectricityConsumption.calculated,
    )
    # Fifth reactor (aerobic) - CSTR with injection
    m.fs.R5 = AerationTank(
        property_package=m.fs.props_ASM1,
        reaction_package=m.fs.ASM1_rxn_props,
        electricity_consumption=ElectricityConsumption.calculated,
    )
    m.fs.SP5 = Separator(
        property_package=m.fs.props_ASM1, outlet_list=["underflow", "overflow"]
    )
    # Clarifier
    # TODO: Replace with more detailed model when available
    m.fs.CL1 = Clarifier(
        property_package=m.fs.props_ASM1,
        outlet_list=["underflow", "effluent"],
        split_basis=SplittingType.componentFlow,
    )
    # Sludge purge splitter
    m.fs.SP6 = Separator(
        property_package=m.fs.props_ASM1,
        outlet_list=["recycle", "waste"],
        split_basis=SplittingType.totalFlow,
    )
    # Mixing sludge recycle and R5 underflow
    m.fs.MX6 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["clarifier", "reactor"],
        momentum_mixing_type=MomentumMixingType.none,
    )

    # Product Blocks
    m.fs.Treated = Product(property_package=m.fs.props_ASM1)
    m.fs.Sludge = Product(property_package=m.fs.props_ASM1)
    # Recycle pressure changer - use a simple isothermal unit for now
    m.fs.P1 = PressureChanger(property_package=m.fs.props_ASM1)

    # Link units
    m.fs.stream2 = Arc(source=m.fs.MX1.outlet, destination=m.fs.R1.inlet)
    m.fs.stream3 = Arc(source=m.fs.R1.outlet, destination=m.fs.R2.inlet)
    m.fs.stream4 = Arc(source=m.fs.R2.outlet, destination=m.fs.R3.inlet)
    m.fs.stream5 = Arc(source=m.fs.R3.outlet, destination=m.fs.R4.inlet)
    m.fs.stream6 = Arc(source=m.fs.R4.outlet, destination=m.fs.R5.inlet)
    m.fs.stream7 = Arc(source=m.fs.R5.outlet, destination=m.fs.SP5.inlet)
    m.fs.stream8 = Arc(source=m.fs.SP5.overflow, destination=m.fs.CL1.inlet)
    m.fs.stream9 = Arc(source=m.fs.SP5.underflow, destination=m.fs.MX6.reactor)
    m.fs.stream10 = Arc(source=m.fs.CL1.effluent, destination=m.fs.Treated.inlet)
    m.fs.stream11 = Arc(source=m.fs.CL1.underflow, destination=m.fs.SP6.inlet)
    m.fs.stream13 = Arc(source=m.fs.SP6.recycle, destination=m.fs.MX6.clarifier)
    m.fs.stream14 = Arc(source=m.fs.MX6.outlet, destination=m.fs.P1.inlet)
    m.fs.stream15 = Arc(source=m.fs.P1.outlet, destination=m.fs.MX1.recycle)
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Oxygen concentration in reactors 3 and 4 is governed by mass transfer
    m.fs.R3.KLa = 7.6
    m.fs.R4.KLa = 5.7

    # ======================================================================
    # Anaerobic digester section
    m.fs.asm_adm = Translator_ASM1_ADM1(
        inlet_property_package=m.fs.props_ASM1,
        outlet_property_package=m.fs.props_ADM1,
        reaction_package=m.fs.ADM1_rxn_props,
        has_phase_equilibrium=False,
        outlet_state_defined=True,
    )

    m.fs.RADM = AD(
        liquid_property_package=m.fs.props_ADM1,
        vapor_property_package=m.fs.props_vap,
        reaction_package=m.fs.ADM1_rxn_props,
        has_heat_transfer=True,
        has_pressure_change=False,
    )

    m.fs.adm_asm = Translator_ADM1_ASM1(
        inlet_property_package=m.fs.props_ADM1,
        outlet_property_package=m.fs.props_ASM1,
        reaction_package=m.fs.ADM1_rxn_props,
        has_phase_equilibrium=False,
        outlet_state_defined=True,
    )

    # ====================================================================
    # Primary Clarifier
    m.fs.CL = Clarifier(
        property_package=m.fs.props_ASM1,
        outlet_list=["underflow", "effluent"],
        split_basis=SplittingType.componentFlow,
    )

    # Thickener
    m.fs.TU = Thickener(property_package=m.fs.props_ASM1)
    # Dewaterer
    m.fs.DU = DewateringUnit(property_package=m.fs.props_ASM1)

    m.fs.MX2 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["feed_water1", "recycle1"],
        momentum_mixing_type=MomentumMixingType.none,
    )
    m.fs.MX3 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["feed_water2", "recycle2"],
        momentum_mixing_type=MomentumMixingType.none,
    )
    m.fs.MX4 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["thickener", "clarifier"],
        momentum_mixing_type=MomentumMixingType.none,
    )

    # Make connections related to AD section
    m.fs.stream2adm = Arc(
        source=m.fs.RADM.liquid_outlet, destination=m.fs.adm_asm.inlet
    )
    m.fs.stream6adm = Arc(source=m.fs.SP6.waste, destination=m.fs.TU.inlet)
    m.fs.stream3adm = Arc(source=m.fs.TU.underflow, destination=m.fs.MX4.thickener)
    m.fs.stream7adm = Arc(source=m.fs.TU.overflow, destination=m.fs.MX3.recycle2)
    m.fs.stream9adm = Arc(source=m.fs.CL.underflow, destination=m.fs.MX4.clarifier)
    m.fs.stream4adm = Arc(source=m.fs.adm_asm.outlet, destination=m.fs.DU.inlet)
    m.fs.stream5adm = Arc(source=m.fs.DU.overflow, destination=m.fs.MX2.recycle1)
    m.fs.stream11adm = Arc(source=m.fs.DU.underflow, destination=m.fs.Sludge.inlet)
    m.fs.stream01 = Arc(source=m.fs.FeedWater.outlet, destination=m.fs.MX2.feed_water1)
    m.fs.stream02 = Arc(source=m.fs.MX2.outlet, destination=m.fs.MX3.feed_water2)
    m.fs.stream03 = Arc(source=m.fs.MX3.outlet, destination=m.fs.CL.inlet)
    m.fs.stream04 = Arc(source=m.fs.CL.effluent, destination=m.fs.MX1.feed_water)
    m.fs.stream10adm = Arc(source=m.fs.MX4.outlet, destination=m.fs.asm_adm.inlet)
    m.fs.stream1adm = Arc(source=m.fs.asm_adm.outlet, destination=m.fs.RADM.inlet)

    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # keep handy all the mixers
    m.fs.mixers = (m.fs.MX1, m.fs.MX2, m.fs.MX3, m.fs.MX4, m.fs.MX6)

    return m


def set_operating_conditions(m):
    # Feed Water Conditions
    m.fs.FeedWater.flow_vol.fix(20648 * pyo.units.m**3 / pyo.units.day)
    m.fs.FeedWater.temperature.fix(308.15 * pyo.units.K)
    m.fs.FeedWater.pressure.fix(1 * pyo.units.atm)
    m.fs.FeedWater.conc_mass_comp[0, "S_I"].fix(27 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_S"].fix(58 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_I"].fix(92 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_S"].fix(363 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_BH"].fix(50 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_BA"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_P"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_O"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NO"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NH"].fix(23 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_ND"].fix(5 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_ND"].fix(16 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.alkalinity.fix(7 * pyo.units.mol / pyo.units.m**3)

    # Reactor sizing in activated sludge process
    m.fs.R1.volume.fix(1000 * pyo.units.m**3)
    m.fs.R2.volume.fix(1000 * pyo.units.m**3)
    m.fs.R3.volume.fix(1333 * pyo.units.m**3)
    m.fs.R4.volume.fix(1333 * pyo.units.m**3)
    m.fs.R5.volume.fix(1333 * pyo.units.m**3)

    # Injection rates to Reactors 3, 4 and 5 of the activated sludge process
    for j in m.fs.props_ASM1.component_list:
        if j != "S_O":
            # All components except S_O have no injection
            m.fs.R3.injection[:, :, j].fix(0)
            m.fs.R4.injection[:, :, j].fix(0)
            m.fs.R5.injection[:, :, j].fix(0)
    # Then set injections rates for O2
    m.fs.R3.outlet.conc_mass_comp[:, "S_O"].fix(1.72e-3)
    m.fs.R4.outlet.conc_mass_comp[:, "S_O"].fix(2.43e-3)
    m.fs.R5.outlet.conc_mass_comp[:, "S_O"].fix(4.49e-4)

    # Set fraction of outflow from reactor 5 that goes to recycle
    m.fs.SP5.split_fraction[:, "underflow"].fix(0.6)

    # Secondary clarifier
    # TODO: Update once secondary clarifier with more detailed model available
    m.fs.CL1.split_fraction[0, "effluent", "H2O"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_I"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_S"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "X_I"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_S"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_BH"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_BA"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_P"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "S_O"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NO"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NH"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_ND"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "X_ND"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "S_ALK"].fix(0.48956)

    m.fs.CL1.surface_area.fix(1500 * pyo.units.m**2)

    # Sludge purge separator
    m.fs.SP6.split_fraction[:, "recycle"].fix(0.985)

    # Outlet pressure from recycle pump
    m.fs.P1.outlet.pressure.fix(101325)

    # Primary Clarifier
    # TODO: Update primary clarifier once more detailed model available
    m.fs.CL.split_fraction[0, "effluent", "H2O"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_I"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_S"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "X_I"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_S"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_BH"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_BA"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_P"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "S_O"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_NO"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_NH"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_ND"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "X_ND"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "S_ALK"].fix(0.993)

    # Anaerobic digester
    m.fs.RADM.volume_liquid.fix(3400)
    m.fs.RADM.volume_vapor.fix(300)
    m.fs.RADM.liquid_outlet.temperature.fix(308.15)

    # Dewatering Unit - fix either HRT or volume.
    m.fs.DU.hydraulic_retention_time.fix(1800 * pyo.units.s)

    # Set specific energy consumption averaged for centrifuge
    m.fs.DU.energy_electric_flow_vol_inlet[0] = 0.069 * pyo.units.kWh / pyo.units.m**3

    # Thickener unit
    m.fs.TU.hydraulic_retention_time.fix(86400 * pyo.units.s)
    m.fs.TU.diameter.fix(10 * pyo.units.m)

    # Mixers - fix outlet pressures since MomentumMixingType.none and isobaric assumption
    for mixer in m.fs.mixers:
        mixer.outlet.pressure.fix()


def scale_system(m):
    m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

    csb = CustomScalerBase()

    ad_scaler = ADScaler()
    ad_scaler.scale_model(m.fs.RADM)
    # Poorly scaled Jacobians
    set_scaling_factor(m.fs.RADM.liquid_phase.heat[0], 1e-3, overwrite=True)
    set_scaling_factor(
        m.fs.RADM.liquid_phase.enthalpy_transfer[0], 1e-2, overwrite=True
    )
    set_scaling_factor(m.fs.RADM.liquid_phase.reactions[0].S_H, 1e7)

    for c in m.fs.props_vap.solute_set:
        set_scaling_factor(
            m.fs.RADM.vapor_phase[0].conc_mass_comp[c], 1e2, overwrite=True
        )
    set_scaling_factor(
        m.fs.RADM.vapor_phase[0].conc_mass_comp["S_h2"], 1e3, overwrite=True
    )

    for c in m.fs.props_vap.component_list:
        set_scaling_factor(
            m.fs.RADM.vapor_phase[0].pressure_sat[c], 1e-3, overwrite=True
        )
    set_scaling_factor(
        m.fs.RADM.liquid_phase.properties_in[0].conc_mass_comp["X_aa"],
        1,
        overwrite=True,
    )
    set_scaling_factor(
        m.fs.RADM.liquid_phase.properties_in[0].conc_mass_comp["X_h2"],
        1,
        overwrite=True,
    )

    set_scaling_factor(m.fs.RADM.KH_h2[0], 1e4)
    set_scaling_factor(
        m.fs.RADM.KH_ch4[0], 1e3
    )  # NOTE: This drastically reduced the number of iterations
    for c in m.fs.props_ADM1.solute_set:
        set_scaling_factor(
            m.fs.RADM.liquid_phase.mass_transfer_term[0, "Liq", c],
            1e10,
            overwrite=True,
        )
    set_scaling_factor(
        m.fs.RADM.liquid_phase.mass_transfer_term[0, "Liq", "H2O"],
        1e4,
        overwrite=True,
    )
    set_scaling_factor(
        m.fs.RADM.liquid_phase.mass_transfer_term[0, "Liq", "S_h2"],
        1e2,
        overwrite=True,
    )
    set_scaling_factor(
        m.fs.RADM.liquid_phase.mass_transfer_term[0, "Liq", "S_ch4"],
        1e2,
        overwrite=True,
    )
    set_scaling_factor(
        m.fs.RADM.liquid_phase.mass_transfer_term[0, "Liq", "S_cat"],
        1e9,
        overwrite=True,
    )
    set_scaling_factor(
        m.fs.RADM.liquid_phase.mass_transfer_term[0, "Liq", "S_an"],
        1e9,
        overwrite=True,
    )
    set_scaling_factor(
        m.fs.RADM.liquid_phase.mass_transfer_term[0, "Liq", "S_IC"],
        1e3,
        overwrite=True,
    )

    for rxn in m.fs.ADM1_rxn_props.rate_reaction_idx:
        set_scaling_factor(
            m.fs.RADM.liquid_phase.reactions[0].reaction_rate[rxn],
            1e7,
            overwrite=True,
        )

    set_scaling_factor(
        m.fs.RADM.liquid_phase.reactions[0].reaction_rate["R1"],
        1e5,
        overwrite=True,
    )
    set_scaling_factor(
        m.fs.RADM.liquid_phase.reactions[0].reaction_rate["R11"],
        1e5,
        overwrite=True,
    )

    cstr_list = [m.fs.R1, m.fs.R2]
    cstr_scaler = CSTRScaler()
    for unit in cstr_list:
        cstr_scaler.scale_model(unit)

    aeration_list = [m.fs.R3, m.fs.R4, m.fs.R5]
    aeration_scaler = AerationTankScaler()
    for unit in aeration_list:
        aeration_scaler.scale_model(unit)
    set_scaling_factor(m.fs.R3.outlet.conc_mass_comp[0, "S_O"], 1e3)
    set_scaling_factor(m.fs.R4.outlet.conc_mass_comp[0, "S_O"], 1e3)
    set_scaling_factor(m.fs.R5.outlet.conc_mass_comp[0, "S_O"], 1e3)

    reactor_list = [m.fs.R1, m.fs.R2, m.fs.R3, m.fs.R4, m.fs.R5]
    for r in reactor_list:
        set_scaling_factor(
            r.control_volume.properties_in[0].flow_vol,
            1e0,
            overwrite=True,
        )
        set_scaling_factor(
            r.control_volume.properties_out[0].flow_vol,
            1e0,
            overwrite=True,
        )
        set_scaling_factor(
            r.control_volume.properties_in[0].conc_mass_comp["X_BH"],
            1e0,
            overwrite=True,
        )
        set_scaling_factor(
            r.control_volume.properties_in[0].conc_mass_comp["X_BA"],
            1e0,
            overwrite=True,
        )
        set_scaling_factor(
            r.control_volume.properties_in[0].conc_mass_comp["X_P"],
            1e0,
            overwrite=True,
        )

        set_scaling_factor(
            r.control_volume.properties_out[0].conc_mass_comp["X_I"],
            1e0,
            overwrite=True,
        )
        set_scaling_factor(
            r.control_volume.properties_out[0].conc_mass_comp["X_BH"],
            1e0,
            overwrite=True,
        )
        set_scaling_factor(
            r.control_volume.properties_out[0].conc_mass_comp["X_BA"],
            1e0,
            overwrite=True,
        )
        set_scaling_factor(
            r.control_volume.properties_out[0].conc_mass_comp["X_P"],
            1e0,
            overwrite=True,
        )
        set_scaling_factor(
            r.control_volume.properties_out[0].conc_mass_comp["S_O"],
            1e6,
            overwrite=True,
        )

        for c in m.fs.props_ASM1.component_list:
            set_scaling_factor(
                r.control_volume.rate_reaction_generation[0, "Liq", c],
                1e3,
                overwrite=True,
            )

        for rxn in m.fs.ASM1_rxn_props.rate_reaction_idx:
            set_scaling_factor(
                r.control_volume.rate_reaction_extent[0, rxn],
                1e3,
                overwrite=True,
            )
            set_scaling_factor(
                r.control_volume.reactions[0].reaction_rate[rxn],
                1e6,
                overwrite=True,
            )

    clarifier_list = [m.fs.CL, m.fs.CL1]
    clarifier_scaler = ClarifierScaler()
    for unit in clarifier_list:
        clarifier_scaler.scale_model(unit)

    thickener_scaler = ThickenerScaler()
    thickener_scaler.scale_model(m.fs.TU)

    dewaterer_scaler = DewatererScaler()
    dewaterer_scaler.scale_model(m.fs.DU)

    as_ad_scaler = ASM1ADM1Scaler()
    as_ad_scaler.scale_model(m.fs.asm_adm)

    ad_as_scaler = ADM1ASM1Scaler()
    ad_as_scaler.scale_model(m.fs.adm_asm)

    set_scaling_factor(m.fs.MX1.feed_water_state[0].conc_mass_comp["S_I"], 1e1)
    set_scaling_factor(m.fs.MX1.feed_water_state[0].conc_mass_comp["X_S"], 1e2)
    set_scaling_factor(m.fs.MX1.feed_water_state[0].conc_mass_comp["S_O"], 1e6)

    set_scaling_factor(m.fs.MX1.recycle_state[0].conc_mass_comp["X_I"], 1e1)
    set_scaling_factor(m.fs.MX1.recycle_state[0].conc_mass_comp["X_BA"], 1e1)
    set_scaling_factor(m.fs.MX1.recycle_state[0].conc_mass_comp["X_P"], 1e1)

    set_scaling_factor(m.fs.MX1.mixed_state[0].flow_vol, 1e0)
    set_scaling_factor(m.fs.MX1.mixed_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.MX1.mixed_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.MX1.mixed_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.MX1.mixed_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.SP5.mixed_state[0].flow_vol, 1e1)
    set_scaling_factor(m.fs.SP5.mixed_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.SP5.mixed_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.SP5.mixed_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.SP5.mixed_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.SP5.underflow_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.SP5.underflow_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.SP5.underflow_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.SP5.underflow_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.SP5.overflow_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.SP5.overflow_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.SP5.overflow_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.SP5.overflow_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.CL1.mixed_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.CL1.mixed_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.CL1.mixed_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.CL1.mixed_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.CL1.underflow_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.CL1.underflow_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.CL1.underflow_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.CL1.underflow_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.CL1.underflow_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.SP6.mixed_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.SP6.mixed_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.SP6.mixed_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.SP6.mixed_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.SP6.mixed_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.SP6.recycle_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.SP6.recycle_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.SP6.recycle_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.SP6.recycle_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.SP6.recycle_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.SP6.waste_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.SP6.waste_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.SP6.waste_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.SP6.waste_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.SP6.waste_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.MX6.mixed_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.MX6.mixed_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.MX6.mixed_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.MX6.mixed_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.MX6.reactor_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.MX6.reactor_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.MX6.reactor_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.MX6.reactor_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.MX6.clarifier_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.MX6.clarifier_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.MX6.clarifier_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.MX6.clarifier_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.MX6.clarifier_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.MX4.mixed_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.MX4.mixed_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.MX4.mixed_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.MX4.mixed_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.MX4.mixed_state[0].conc_mass_comp["X_P"], 1e0)
    set_scaling_factor(m.fs.MX4.mixed_state[0].conc_mass_comp["X_ND"], 1e0)

    set_scaling_factor(m.fs.MX4.thickener_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.MX4.thickener_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.MX4.thickener_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.MX4.thickener_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.MX4.thickener_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.MX4.clarifier_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.MX4.clarifier_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.MX4.clarifier_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.MX4.clarifier_state[0].conc_mass_comp["S_O"], 1e6)
    set_scaling_factor(m.fs.MX4.clarifier_state[0].conc_mass_comp["X_ND"], 1e0)

    set_scaling_factor(m.fs.MX3.feed_water2_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.MX3.feed_water2_state[0].conc_mass_comp["X_BA"], 1e8)
    set_scaling_factor(m.fs.MX3.feed_water2_state[0].conc_mass_comp["X_P"], 1e8)
    set_scaling_factor(m.fs.MX3.feed_water2_state[0].conc_mass_comp["S_O"], 1e8)
    set_scaling_factor(m.fs.MX3.feed_water2_state[0].conc_mass_comp["S_NO"], 1e8)
    set_scaling_factor(m.fs.MX3.recycle2_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.MX3.mixed_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.MX3.mixed_state[0].conc_mass_comp["S_O"], 1e6)

    set_scaling_factor(m.fs.MX2.recycle1_state[0].conc_mass_comp["S_I"], 1e0)
    set_scaling_factor(m.fs.MX2.recycle1_state[0].conc_mass_comp["S_S"], 1e0)
    set_scaling_factor(m.fs.MX2.recycle1_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.MX2.recycle1_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.MX2.recycle1_state[0].conc_mass_comp["S_O"], 1e10)
    set_scaling_factor(m.fs.MX2.recycle1_state[0].conc_mass_comp["S_NO"], 1e10)
    set_scaling_factor(m.fs.MX2.recycle1_state[0].conc_mass_comp["S_NH"], 1e0)
    set_scaling_factor(m.fs.MX2.recycle1_state[0].conc_mass_comp["S_ND"], 1e0)

    set_scaling_factor(m.fs.MX2.feed_water1_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.MX2.feed_water1_state[0].conc_mass_comp["X_BA"], 1e8)
    set_scaling_factor(m.fs.MX2.feed_water1_state[0].conc_mass_comp["X_P"], 1e8)
    set_scaling_factor(m.fs.MX2.feed_water1_state[0].conc_mass_comp["S_O"], 1e8)
    set_scaling_factor(m.fs.MX2.feed_water1_state[0].conc_mass_comp["S_NO"], 1e8)

    set_scaling_factor(m.fs.MX2.mixed_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.MX2.mixed_state[0].conc_mass_comp["X_BA"], 1e8)
    set_scaling_factor(m.fs.MX2.mixed_state[0].conc_mass_comp["X_P"], 1e8)
    set_scaling_factor(m.fs.MX2.mixed_state[0].conc_mass_comp["S_O"], 1e8)
    set_scaling_factor(m.fs.MX2.mixed_state[0].conc_mass_comp["S_NO"], 1e8)

    set_scaling_factor(m.fs.DU.mixed_state[0].flow_vol, 1e3, overwrite=True)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["S_I"], 1e0)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["S_S"], 1e0)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["X_I"], 1e-2)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["X_S"], 1e-2)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["S_NH"], 1e0)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["S_ND"], 1e0)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["X_ND"], 1e0)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["X_BH"], 1e10)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["S_O"], 1e10)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["S_NO"], 1e10)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["X_BA"], 1e10)
    set_scaling_factor(m.fs.DU.mixed_state[0].conc_mass_comp["X_P"], 1e10)

    set_scaling_factor(m.fs.DU.underflow_state[0].flow_vol, 1e5, overwrite=True)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["S_I"], 1e0)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["S_S"], 1e0)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["X_I"], 1e-2)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["X_S"], 1e-2)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["S_NH"], 1e0)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["S_ND"], 1e0)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["X_ND"], 1e0)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["X_BH"], 1e10)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["S_O"], 1e10)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["S_NO"], 1e10)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["X_BA"], 1e10)
    set_scaling_factor(m.fs.DU.underflow_state[0].conc_mass_comp["X_P"], 1e10)

    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["S_I"], 1e0)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["S_S"], 1e0)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["S_NH"], 1e0)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["S_ND"], 1e0)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["X_BH"], 1e10)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["S_O"], 1e10)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["S_NO"], 1e10)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["X_BA"], 1e10)
    set_scaling_factor(m.fs.DU.overflow_state[0].conc_mass_comp["X_P"], 1e10)

    set_scaling_factor(m.fs.TU.underflow_state[0].flow_vol, 1e4, overwrite=True)
    set_scaling_factor(m.fs.TU.underflow_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.TU.underflow_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.TU.underflow_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.TU.underflow_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.TU.underflow_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.TU.mixed_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.TU.mixed_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.TU.mixed_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.TU.mixed_state[0].conc_mass_comp["X_BA"], 1e0)
    set_scaling_factor(m.fs.TU.mixed_state[0].conc_mass_comp["X_P"], 1e0)

    set_scaling_factor(m.fs.TU.overflow_state[0].conc_mass_comp["X_BH"], 1e0)

    set_scaling_factor(m.fs.CL.effluent_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.CL.effluent_state[0].conc_mass_comp["S_O"], 1e6)

    set_scaling_factor(m.fs.CL.mixed_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.CL.mixed_state[0].conc_mass_comp["S_O"], 1e6)

    set_scaling_factor(m.fs.CL.underflow_state[0].conc_mass_comp["X_I"], 1e0)
    set_scaling_factor(m.fs.CL.underflow_state[0].conc_mass_comp["X_S"], 1e0)
    set_scaling_factor(m.fs.CL.underflow_state[0].conc_mass_comp["X_BH"], 1e0)
    set_scaling_factor(m.fs.CL.underflow_state[0].conc_mass_comp["X_ND"], 1e0)
    set_scaling_factor(m.fs.CL.underflow_state[0].conc_mass_comp["S_O"], 1e6)

    set_scaling_factor(m.fs.P1.control_volume.work[0], 1e-4)
    set_scaling_factor(m.fs.P1.control_volume.deltaP[0], 1e-3)
    set_scaling_factor(
        m.fs.P1.control_volume.properties_in[0].conc_mass_comp["X_I"], 1e0
    )
    set_scaling_factor(
        m.fs.P1.control_volume.properties_in[0].conc_mass_comp["X_BH"], 1e0
    )
    set_scaling_factor(
        m.fs.P1.control_volume.properties_in[0].conc_mass_comp["X_BA"], 1e0
    )
    set_scaling_factor(
        m.fs.P1.control_volume.properties_in[0].conc_mass_comp["X_P"], 1e0
    )

    set_scaling_factor(
        m.fs.P1.control_volume.properties_out[0].conc_mass_comp["X_I"], 1e0
    )
    set_scaling_factor(
        m.fs.P1.control_volume.properties_out[0].conc_mass_comp["X_BH"], 1e0
    )
    set_scaling_factor(
        m.fs.P1.control_volume.properties_out[0].conc_mass_comp["X_BA"], 1e0
    )
    set_scaling_factor(
        m.fs.P1.control_volume.properties_out[0].conc_mass_comp["X_P"], 1e0
    )

    set_scaling_factor(m.fs.Sludge.properties[0].flow_vol, 1e5)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["S_I"], 1e0)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["S_S"], 1e0)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["X_I"], 1e-2)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["X_S"], 1e-2)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["X_BH"], 1e9)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["X_BA"], 1e9)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["X_P"], 1e9)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["S_O"], 1e9)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["S_NO"], 1e9)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["S_NH"], 1e0)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["S_ND"], 1e0)
    set_scaling_factor(m.fs.Sludge.properties[0].conc_mass_comp["X_ND"], 1e0)

    for c in m.fs.props_ASM1.component_list:
        if c in ["X_I", "X_S", "X_BH", "X_BA", "X_P", "X_ND"]:
            set_scaling_factor(
                m.fs.CL.split_fraction[0, "underflow", c],
                1e1,
                overwrite=True,
            )
        else:
            set_scaling_factor(
                m.fs.CL.split_fraction[0, "underflow", c],
                1e3,
                overwrite=True,
            )

    for c in m.fs.props_ADM1.solute_set:
        set_scaling_factor(
            m.fs.adm_asm.properties_in[0].conc_mass_comp[c],
            1e0,
            overwrite=True,
        )
    for c in m.fs.props_ASM1.solute_set:
        set_scaling_factor(
            m.fs.adm_asm.properties_out[0].conc_mass_comp[c],
            1e0,
            overwrite=True,
        )
        set_scaling_factor(
            m.fs.asm_adm.properties_in[0].conc_mass_comp[c],
            1e0,
            overwrite=True,
        )
    set_scaling_factor(m.fs.adm_asm.properties_in[0].flow_vol, 1e3, overwrite=True)
    set_scaling_factor(m.fs.adm_asm.properties_out[0].flow_vol, 1e3, overwrite=True)
    set_scaling_factor(
        m.fs.adm_asm.properties_in[0].conc_mass_comp["S_h2"], 1e3, overwrite=True
    )
    set_scaling_factor(
        m.fs.adm_asm.properties_in[0].conc_mass_comp["S_aa"], 1e3, overwrite=True
    )

    set_scaling_factor(
        m.fs.adm_asm.properties_out[0].conc_mass_comp["X_BH"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.adm_asm.properties_out[0].conc_mass_comp["X_BA"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.adm_asm.properties_out[0].conc_mass_comp["X_P"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.adm_asm.properties_out[0].conc_mass_comp["S_O"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.adm_asm.properties_out[0].conc_mass_comp["S_NO"], 1e10, overwrite=True
    )
    set_scaling_factor(m.fs.asm_adm.properties_in[0].flow_vol, 1e3, overwrite=True)
    set_scaling_factor(
        m.fs.asm_adm.properties_in[0].conc_mass_comp["S_O"], 1e6, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_in[0].conc_mass_comp["S_NO"], 1e3, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_in[0].conc_mass_comp["S_ND"], 1e3, overwrite=True
    )

    set_scaling_factor(m.fs.asm_adm.properties_out[0].flow_vol, 1e3, overwrite=True)
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["S_fa"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["S_va"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["S_bu"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["S_pro"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["S_ac"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["S_h2"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["S_ch4"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["S_IN"], 1e0, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_c"], 1e0, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_pr"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_su"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_aa"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_fa"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_c4"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_pro"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_ac"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_h2"], 1e10, overwrite=True
    )
    set_scaling_factor(
        m.fs.asm_adm.properties_out[0].conc_mass_comp["X_I"], 1e0, overwrite=True
    )

    for var in m.fs.component_data_objects(pyo.Var, descend_into=True):
        if "flow_vol" in var.name:
            set_scaling_factor(var, 1e2)
        if "temperature" in var.name:
            set_scaling_factor(var, 1e-2)
        if "pressure" in var.name:
            set_scaling_factor(var, 1e-5)
        if "conc_mass_comp" in var.name:
            set_scaling_factor(var, 1e3)
        if "conc_mol" in var.name:
            set_scaling_factor(var, 1e2)
        if "alkalinity" in var.name:
            set_scaling_factor(var, 1e3)
        if "split_fraction" in var.name:
            set_scaling_factor(var, 1e1)

    for c in m.fs.component_data_objects(pyo.Constraint, descend_into=True):
        csb.scale_constraint_by_nominal_value(
            c,
            scheme=ConstraintScalingScheme.inverseMaximum,
            overwrite=True,
        )


def initialize_system(m):
    # Initialize flowsheet
    # Apply sequential decomposition - 1 iteration should suffice
    seq = SequentialDecomposition()
    seq.options.tear_method = "Direct"
    seq.options.iterLim = 1
    seq.options.tear_set = [m.fs.stream2, m.fs.stream10adm]

    G = seq.create_graph(m)
    # The code below shows tear set and initialization order
    order = seq.calculation_order(G)
    print("Initialization Order")
    for o in order:
        print(o[0].name)

    # Initial guesses for flow into first reactor
    tear_guesses1 = {
        "flow_vol": {0: 103531 / 24 / 3600},
        "conc_mass_comp": {
            (0, "S_I"): 0.028,
            (0, "S_S"): 0.062,
            (0, "X_I"): 1.532,
            (0, "X_S"): 0.069,
            (0, "X_BH"): 2.233,
            (0, "X_BA"): 0.196,
            (0, "X_P"): 0.964,
            (0, "S_O"): 0.0036,
            (0, "S_NO"): 0.0073,
            (0, "S_NH"): 0.0072,
            (0, "S_ND"): 0.002,
            (0, "X_ND"): 0.0040,
        },
        "alkalinity": {0: 0.00438},
        "temperature": {0: 308.15},
        "pressure": {0: 101325},
    }

    tear_guesses2 = {
        "flow_vol": {0: 0.002},
        "conc_mass_comp": {
            (0, "S_I"): 0.028,
            (0, "S_S"): 0.2,
            (0, "X_I"): 9.03,
            (0, "X_S"): 18.0,
            (0, "X_BH"): 10.210,
            (0, "X_BA"): 0.553,
            (0, "X_P"): 3.204,
            (0, "S_O"): 0.00012,
            (0, "S_NO"): 0.00169,
            (0, "S_NH"): 0.0289,
            (0, "S_ND"): 0.00468,
            (0, "X_ND"): 0.906,
        },
        "alkalinity": {0: 0.006},
        "temperature": {0: 308.15},
        "pressure": {0: 101325},
    }

    # Pass the tear_guess to the SD tool
    seq.set_guesses_for(m.fs.R1.inlet, tear_guesses1)
    seq.set_guesses_for(m.fs.asm_adm.inlet, tear_guesses2)

    initializer = BlockTriangularizationInitializer(
        calculate_variable_options={"eps": 2e-8}, skip_final_solve=True
    )

    def function(unit):
        try:
            initializer.initialize(unit, output_level=_log.debug)
        except InitializationError:
            pass

    seq.run(m, function)


def add_costing(m):
    m.fs.costing = WaterTAPCosting()
    m.fs.costing.base_currency = pyo.units.USD_2020

    # Costing Blocks
    m.fs.R1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R2.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R3.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R4.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R5.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.CL.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_primary_clarifier,
    )

    m.fs.CL1.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_circular_clarifier,
    )

    m.fs.RADM.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.DU.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.TU.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # Initialize electricity consumption values
    m.fs.R3.electricity_consumption[0].set_value(75)
    m.fs.R4.electricity_consumption[0].set_value(70)
    m.fs.R5.electricity_consumption[0].set_value(20)
    m.fs.CL.electricity_consumption[0].set_value(10)
    m.fs.CL1.electricity_consumption[0].set_value(15)
    m.fs.RADM.electricity_consumption[0].set_value(25)
    m.fs.TU.electricity_consumption[0].set_value(1)
    m.fs.DU.electricity_consumption[0].set_value(1)

    # process costing and add system level metrics
    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(m.fs.Treated.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.FeedWater.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(m.fs.FeedWater.properties[0].flow_vol)

    m.fs.objective = pyo.Objective(expr=m.fs.costing.LCOW)

    set_scaling_factor(m.fs.costing.total_capital_cost, 1e-7)
    set_scaling_factor(m.fs.costing.aggregate_capital_cost, 1e-8)
    set_scaling_factor(m.fs.costing.aggregate_flow_electricity, 1e-1)
    set_scaling_factor(m.fs.costing.aggregate_flow_costs["electricity"], 1e-3)
    set_scaling_factor(m.fs.costing.total_operating_cost, 1e-4)

    for block in m.fs.component_objects(pyo.Block, descend_into=True):
        if isinstance(block, UnitModelBlockData) and hasattr(block, "costing"):
            set_scaling_factor(block.costing.capital_cost, 1e-7)


def setup_optimization(m, reactor_volume_equalities=True):

    for i in ["R1", "R2", "R3", "R4", "R5"]:
        reactor = getattr(m.fs, i)
        reactor.volume.unfix()
        reactor.volume.setlb(1)
        reactor.volume.setub(2000)
    if reactor_volume_equalities:
        add_reactor_volume_equalities(m)

    m.fs.R3.outlet.conc_mass_comp[:, "S_O"].unfix()
    m.fs.R3.outlet.conc_mass_comp[:, "S_O"].setub(8.2e-3)

    m.fs.R4.outlet.conc_mass_comp[:, "S_O"].unfix()
    m.fs.R4.outlet.conc_mass_comp[:, "S_O"].setub(8.2e-3)

    m.fs.R5.outlet.conc_mass_comp[:, "S_O"].unfix()
    m.fs.R5.outlet.conc_mass_comp[:, "S_O"].setub(8.2e-3)

    # Unfix fraction of outflow from reactor 5 that goes to recycle
    m.fs.SP5.split_fraction[:, "underflow"].unfix()
    # m.fs.SP5.split_fraction[:, "underflow"].setlb(0.45)
    m.fs.SP6.split_fraction[:, "recycle"].unfix()

    add_effluent_violations(m)


def add_effluent_violations(m):
    # TODO: update "m" to blk; change ref to m.fs.Treated instead of CL1 effluent
    m.fs.TSS_max = pyo.Var(initialize=0.03, units=pyo.units.kg / pyo.units.m**3)
    m.fs.TSS_max.fix()

    @m.fs.Constraint(m.fs.time)
    def eq_TSS_max(self, t):
        return m.fs.CL1.effluent_state[0].TSS <= m.fs.TSS_max

    m.fs.COD_max = pyo.Var(initialize=0.1, units=pyo.units.kg / pyo.units.m**3)
    m.fs.COD_max.fix()

    @m.fs.Constraint(m.fs.time)
    def eq_COD_max(self, t):
        return m.fs.CL1.effluent_state[0].COD <= m.fs.COD_max

    m.fs.totalN_max = pyo.Var(initialize=0.018, units=pyo.units.kg / pyo.units.m**3)
    m.fs.totalN_max.fix()

    @m.fs.Constraint(m.fs.time)
    def eq_totalN_max(self, t):
        return m.fs.CL1.effluent_state[0].Total_N <= m.fs.totalN_max

    m.fs.BOD5_max = pyo.Var(initialize=0.01, units=pyo.units.kg / pyo.units.m**3)
    m.fs.BOD5_max.fix()

    @m.fs.Constraint(m.fs.time)
    def eq_BOD5_max(self, t):
        return m.fs.CL1.effluent_state[0].BOD5["effluent"] <= m.fs.BOD5_max


def add_reactor_volume_equalities(m):
    # TODO: These constraints were applied for initial optimization of AS reactor volumes; otherwise, volumes drive towards lower bound. Revisit
    @m.fs.Constraint(m.fs.time)
    def Vol_1(self, t):
        return m.fs.R1.volume[0] == m.fs.R2.volume[0]

    @m.fs.Constraint(m.fs.time)
    def Vol_2(self, t):
        return m.fs.R3.volume[0] == m.fs.R4.volume[0]

    @m.fs.Constraint(m.fs.time)
    def Vol_3(self, t):
        return m.fs.R5.volume[0] >= m.fs.R4.volume[0] * 0.5


def rescale_system(m):
    csb = CustomScalerBase()

    # Scales the new constraints from setup_optimization
    for c in m.fs.component_data_objects(pyo.Constraint, descend_into=True):
        csb.scale_constraint_by_nominal_value(
            c,
            scheme=ConstraintScalingScheme.inverseMaximum,
            overwrite=True,
        )


def solve(blk, solver=None, tee=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    return results


def display_results(m):
    m.display()

    unit_list = [
        "FeedWater",
        "MX1",
        "R1",
        "R2",
        "R3",
        "R4",
        "R5",
        "SP5",
        "CL1",
        "SP6",
        "MX6",
        "Treated",
        "Sludge",
        "P1",
        "asm_adm",
        "RADM",
        "adm_asm",
        "CL",
        "TU",
        "DU",
        "MX2",
        "MX3",
        "MX4",
    ]
    for u in unit_list:
        m.fs.component(u).report()


def display_costing(m):
    print("Levelized cost of water: %.2f $/m3" % pyo.value(m.fs.costing.LCOW))

    print(
        "Total operating cost: %.2f $/yr" % pyo.value(m.fs.costing.total_operating_cost)
    )
    print("Total capital cost: %.2f $" % pyo.value(m.fs.costing.total_capital_cost))

    print(
        "Total annualized cost: %.2f $/yr"
        % pyo.value(m.fs.costing.total_annualized_cost)
    )
    print(
        "capital cost R1",
        pyo.value(m.fs.R1.costing.capital_cost),
        pyo.units.get_units(m.fs.R1.costing.capital_cost),
    )
    print(
        "capital cost R2",
        pyo.value(m.fs.R2.costing.capital_cost),
        pyo.units.get_units(m.fs.R2.costing.capital_cost),
    )
    print(
        "capital cost R3",
        pyo.value(m.fs.R3.costing.capital_cost),
        pyo.units.get_units(m.fs.R3.costing.capital_cost),
    )
    print(
        "capital cost R4",
        pyo.value(m.fs.R4.costing.capital_cost),
        pyo.units.get_units(m.fs.R4.costing.capital_cost),
    )
    print(
        "capital cost R5",
        pyo.value(m.fs.R5.costing.capital_cost),
        pyo.units.get_units(m.fs.R5.costing.capital_cost),
    )
    print(
        "capital cost primary clarifier",
        pyo.value(m.fs.CL.costing.capital_cost),
        pyo.units.get_units(m.fs.CL.costing.capital_cost),
    )
    print(
        "capital cost secondary clarifier",
        pyo.value(m.fs.CL1.costing.capital_cost),
        pyo.units.get_units(m.fs.CL1.costing.capital_cost),
    )
    print(
        "capital cost AD",
        pyo.value(m.fs.RADM.costing.capital_cost),
        pyo.units.get_units(m.fs.RADM.costing.capital_cost),
    )
    print(
        "capital cost dewatering Unit",
        pyo.value(m.fs.DU.costing.capital_cost),
        pyo.units.get_units(m.fs.DU.costing.capital_cost),
    )
    print(
        "capital cost thickener unit",
        pyo.value(m.fs.TU.costing.capital_cost),
        pyo.units.get_units(m.fs.TU.costing.capital_cost),
    )


def display_performance_metrics(m):
    print("Levelized cost of water: %.1f $/m3" % pyo.value(m.fs.costing.LCOW))
    print(
        "Specific energy consumption with respect to influent flowrate: %.1f kWh/m3"
        % pyo.value(m.fs.costing.specific_energy_consumption)
    )

    print(
        "electricity consumption R3",
        pyo.value(m.fs.R3.electricity_consumption[0]),
        pyo.units.get_units(m.fs.R3.electricity_consumption[0]),
    )
    print(
        "electricity consumption R4",
        pyo.value(m.fs.R4.electricity_consumption[0]),
        pyo.units.get_units(m.fs.R4.electricity_consumption[0]),
    )
    print(
        "electricity consumption R5",
        pyo.value(m.fs.R5.electricity_consumption[0]),
        pyo.units.get_units(m.fs.R5.electricity_consumption[0]),
    )
    print(
        "electricity consumption primary clarifier",
        pyo.value(m.fs.CL.electricity_consumption[0]),
        pyo.units.get_units(m.fs.CL.electricity_consumption[0]),
    )
    print(
        "electricity consumption secondary clarifier",
        pyo.value(m.fs.CL1.electricity_consumption[0]),
        pyo.units.get_units(m.fs.CL1.electricity_consumption[0]),
    )
    print(
        "electricity consumption AD",
        pyo.value(m.fs.RADM.electricity_consumption[0]),
        pyo.units.get_units(m.fs.RADM.electricity_consumption[0]),
    )
    print(
        "electricity consumption dewatering Unit",
        pyo.value(m.fs.DU.electricity_consumption[0]),
        pyo.units.get_units(m.fs.DU.electricity_consumption[0]),
    )
    print(
        "electricity consumption thickening Unit",
        pyo.value(m.fs.TU.electricity_consumption[0]),
        pyo.units.get_units(m.fs.TU.electricity_consumption[0]),
    )
    print(
        "Influent flow",
        pyo.value(m.fs.FeedWater.flow_vol[0]),
        pyo.units.get_units(m.fs.FeedWater.flow_vol[0]),
    )
    print(
        "flow into R3",
        pyo.value(m.fs.R3.control_volume.properties_in[0].flow_vol),
        pyo.units.get_units(m.fs.R3.control_volume.properties_in[0].flow_vol),
    )
    print(
        "flow into RADM",
        pyo.value(m.fs.RADM.liquid_phase.properties_in[0].flow_vol),
        pyo.units.get_units(m.fs.RADM.liquid_phase.properties_in[0].flow_vol),
    )


if __name__ == "__main__":
    m, results, sm = main(reactor_volume_equalities=True)



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Simple zero-dimensional Nanofiltration unit model.
"""

from pyomo.environ import Block, Constraint, Set, value, Var
from pyomo.common.config import ConfigDict, ConfigValue, In

from idaes.core import (
    declare_process_block_class,
    EnergyBalanceType,
    MomentumBalanceType,
    useDefault,
    UnitModelBlockData,
    MaterialFlowBasis,
)
from idaes.core.initialization import ModularInitializerBase
from idaes.core.scaling import CustomScalerBase, ConstraintScalingScheme
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog


_log = idaeslog.getLogger(__name__)


class Nanofiltration0DInitializer(ModularInitializerBase):
    """
    Initializer for 0D Nanofiltration models.

    """

    CONFIG = ModularInitializerBase.CONFIG()

    def initialize_main_model(
        self,
        model: Block,
    ):
        """
        Initialization routine for main Nanofiltration0D models.

        Args:
            model: Pyomo Block to be initialized.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not.
                Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.

        Returns:
            Pyomo solver results object.

        """
        # Get loggers
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver = self._get_solver()

        # ---------------------------------------------------------------------
        # Initialize inlet state
        prop_init = self.get_submodel_initializer(model.properties_in)

        if prop_init is not None:
            prop_init.initialize(
                model=model.properties_in,
                output_level=self.get_output_level(),
            )
        else:
            raise ValueError(
                "No Initializer found for property package. Please set provide "
                "sub-model initializers or assign a default Initializer."
            )

        init_log.info_high("Initialization Step 1 (Inlet State) Complete.")

        # ---------------------------------------------------------------------
        # Estimate outlet states from inlet and initialize
        for t, in_state in model.properties_in.items():
            state_vars = in_state.define_state_vars()
            state_vars_ret = model.properties_retentate[t].define_state_vars()
            state_vars_per = model.properties_permeate[t].define_state_vars()

            for n, sv in state_vars.items():
                sv_ret = state_vars_ret[n]
                sv_per = state_vars_per[n]

                if sv.local_name == "pressure" and not sv_ret.fixed:
                    # For pressure, only retentate is linked to inlet
                    if hasattr(model, "deltaP"):
                        sv_ret.set_value(sv + model.deltaP[t])
                    else:
                        sv_ret.set_value(sv)
                elif "flow" in sv.local_name:
                    self._init_flow(model, t, in_state, sv, sv_ret, sv_per)
                elif any(
                    sv.local_name.startswith(i) for i in ["mass_frac", "mole_frac"]
                ):
                    self._init_frac(model, t, in_state, sv, sv_ret, sv_per)
                elif sv.local_name.startswith("conc"):
                    self._init_conc(model, t, in_state, sv, sv_ret, sv_per)
                else:
                    # For everything else, assume outlet similar to inlet
                    for k, svd in sv.items():
                        if not sv_ret[k].fixed:
                            sv_ret[k].set_value(svd)
                        if not sv_per[k].fixed:
                            sv_per[k].set_value(svd)

        prop_init.initialize(
            model=model.properties_retentate,
            output_level=self.get_output_level(),
        )
        prop_init.initialize(
            model=model.properties_permeate,
            output_level=self.get_output_level(),
        )

        init_log.info_high("Initialization Step 2 (Outlet States) Complete.")

        # ---------------------------------------------------------------------
        # Solve full model
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)

        init_log.info("Initialization Completed, {}".format(idaeslog.condition(res)))

        return res

    def _init_conc(self, model, t, in_state, sv, sv_ret, sv_per):
        # Component indexed flow - need to apply split fractions
        for k, svd in sv.items():
            if isinstance(k, str):
                # Single indexing set, component is the index
                j = k
            else:
                # Component is always the last index
                j = k[-1]

            # Determine split based on type
            comp = in_state.params.get_component(j)

            if comp.is_solvent():
                # Assume solvent concentrations unchanged
                sv_ret[k].set_value(svd)
                sv_per[k].set_value(svd)
            else:
                if j == model.config.electroneutrality_ion:
                    # Guess electroneutrality ion will be based on solvent recovery
                    split = model.recovery_solvent[t]
                else:
                    # For initialization, assume density is roughly constant and rejection can be used as split fraction
                    split = 1 - model.rejection_comp[t, j]

                if not sv_ret[k].fixed:
                    sv_ret[k].set_value(svd * (1 - split))
                if not sv_per[k].fixed:
                    sv_per[k].set_value(svd * split)

    def _init_flow(self, model, t, in_state, sv, sv_ret, sv_per):
        # Determine if it is a component flow or not
        if sv.local_name.endswith("_comp"):
            # Component indexed flow - need to apply split fractions
            for k, svd in sv.items():
                if isinstance(k, str):
                    # Single indexing set, component is the index
                    j = k
                else:
                    # Component is always the last index
                    j = k[-1]

                # Determine split based on type
                comp = in_state.params.get_component(j)

                if comp.is_solvent() or j == model.config.electroneutrality_ion:
                    # Assume electroneutrality ion will have similar separation to solvents
                    split = model.recovery_solvent[t]
                else:
                    # For initialization, assume density is roughly constant and rejection can be used as split fraction
                    split = 1 - model.rejection_comp[t, j]

                if not sv_ret[k].fixed:
                    sv_ret[k].set_value(svd * (1 - split))
                if not sv_per[k].fixed:
                    sv_per[k].set_value(svd * split)
        else:
            # Assume a total flow basis, and use solvent recovery
            split = model.recovery_solvent[t]
            for k, svd in sv.items():
                if not sv_ret[k].fixed:
                    sv_ret[k].set_value(svd * (1 - split))
                if not sv_per[k].fixed:
                    sv_per[k].set_value(svd * split)

    def _init_frac(self, model, t, in_state, sv, sv_ret, sv_per):
        # First need to iterate over all indices to collect normalized flows
        # Assume a basis of 1 mass or mole unit
        # Also, only single phase property packages supported, so we only need to
        # worry about the component index
        nom_ret_comp_flow = {}
        nom_per_comp_flow = {}

        for k, svd in sv.items():
            if isinstance(k, str):
                # Single indexing set, component is the index
                j = k
            else:
                # Component is always the last index
                j = k[-1]

            # Determine split based on type
            comp = in_state.params.get_component(j)

            if comp.is_solvent() or j == model.config.electroneutrality_ion:
                # Assume electroneutrality ion will have similar separation to solvents
                split = model.recovery_solvent[t]
            else:
                # For initialization, assume density is roughly constant and rejection can be used as split fraction
                split = 1 - model.rejection_comp[t, j]

            nom_ret_comp_flow[j] = value(svd * (1 - split))
            nom_per_comp_flow[j] = value(svd * split)

        nom_ret_tot_flow = sum(f for f in nom_ret_comp_flow.values())
        nom_per_tot_flow = sum(f for f in nom_per_comp_flow.values())

        # # Next set value based on fraction of nominal total flow
        for k, svd in sv.items():
            if isinstance(k, str):
                # Single indexing set, component is the index
                j = k
            else:
                # Component is always the last index
                j = k[-1]

            if not sv_ret[k].fixed:
                sv_ret[k].set_value(nom_ret_comp_flow[j] / nom_ret_tot_flow)
            if not sv_per[k].fixed:
                sv_per[k].set_value(nom_per_comp_flow[j] / nom_per_tot_flow)


class Nanofiltration0DScaler(CustomScalerBase):
    """
    Scaler class for Nanofiltration0D models.
    """

    DEFAULT_SCALING_FACTORS = {
        "deltaP": 1e-4,
        "rejection_comp": 1e2,
        "recovery_solvent": 10,
    }

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        """
        Apply variable scaling to model.

        Args:
            model: model to be scaled
            overwrite: whether to overwrite existing scaling factors
            submodel_scalers: ComponentMap of Scaler objects to be applied to sub-models

        Returns:
            None

        """
        # Scale inlet state
        self.call_submodel_scaler_method(
            submodel=model.properties_in,
            method="variable_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )

        # Propagate scaling from inlet state
        self.propagate_state_scaling(
            target_state=model.properties_retentate,
            source_state=model.properties_in,
            overwrite=overwrite,
        )
        self.propagate_state_scaling(
            target_state=model.properties_permeate,
            source_state=model.properties_in,
            overwrite=overwrite,
        )

        # Scale remaining states
        self.call_submodel_scaler_method(
            submodel=model.properties_retentate,
            method="variable_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )
        self.call_submodel_scaler_method(
            submodel=model.properties_permeate,
            method="variable_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )

        # Scale remaining variables
        # Variables with default scaling factors
        for n in self.DEFAULT_SCALING_FACTORS.keys():
            c = model.find_component(n)
            for v in c.values():
                self.scale_variable_by_default(v, overwrite=overwrite)

    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        """
        Apply constraint scaling to model.

        Args:
            model: model to be scaled
            overwrite: whether to overwrite existing scaling factors
            submodel_scalers: ComponentMap of Scaler objects to be applied to sub-models

        Returns:
            None

        """
        # Call scaling methods for sub-models
        self.call_submodel_scaler_method(
            submodel=model.properties_in,
            method="constraint_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )
        self.call_submodel_scaler_method(
            submodel=model.properties_retentate,
            method="constraint_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )
        self.call_submodel_scaler_method(
            submodel=model.properties_permeate,
            method="constraint_scaling_routine",
            submodel_scalers=submodel_scalers,
            overwrite=overwrite,
        )

        # Scale remaining constraints
        for c in model.component_data_objects(Constraint, descend_into=False):
            self.scale_constraint_by_nominal_value(
                c,
                scheme=ConstraintScalingScheme.inverseMaximum,
                overwrite=overwrite,
            )


@declare_process_block_class("Nanofiltration0D")
class Nanofiltration0DData(UnitModelBlockData):
    """
    Nanofiltration model.
    """

    default_initializer = Nanofiltration0DInitializer
    default_scaler = Nanofiltration0DScaler

    CONFIG = ConfigDict()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. Product blocks are always steady-state.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Product blocks do not contain holdup, thus this must be
    False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for mixer",
            doc="""Property parameter object used to define property
    calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigDict(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigDict with arguments to be passed to a property
    block(s) and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Retentate momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed
    for the retentate side. Only  MomentumBalanceType.none and
    MomentumBalanceType.pressureTotal (default) are supported.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Retentate pressure change term construction flag",
            doc="""Indicates whether terms for retentate pressure change should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include pressure change terms,
    **False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.isothermal,
            domain=In(EnergyBalanceType),
            description="Indicated what type of energy balance should be constructed.",
            doc="""Indicates what type of momentum balance should be constructed
    for the retentate side. Only  EnergyBalanceType.none and
    EnergyBalanceType.isothermal (default) are supported.""",
        ),
    )
    # TODO: Support multiple electroneutrality ions
    # This needs some rule for prioritizing each ion.
    CONFIG.declare(
        "electroneutrality_ion",
        ConfigValue(
            default="Cl_-",
            domain=str,
            description="Balancing ion to use to maintain electroneutrality",
            doc="""Name of ion to use to maintain electroneutrality in
            permeate stream. If an ion is provided, the split fraction
            for this ion will be solved implicitly to ensure electroneutrality
            is maintained in the permeate stream. If None, user must provide
            spl;it fractions for all monovalent ions and electroneutrality is
            not guaranteed.""",
        ),
    )
    CONFIG.declare(
        "passing_species_list",
        ConfigValue(
            default=None,
            domain=list,
            description="List of solutes with low rejection",
            doc="""List of solute species which pass through the nanofiltration membrane
            with low rejection. If not provided (None), solutes will be classified based
            on their charge, with monovalent and uncharged species considered passing.""",
        ),
    )
    CONFIG.declare(
        "default_passing_rejection",
        ConfigValue(
            default=0.1,
            domain=float,
            description="Default value for rejection of passing species",
            doc="""Default value to be assigned as the rejection fraction for species
            identified as passing through the membrane with low rejection.""",
        ),
    )
    CONFIG.declare(
        "default_excluded_rejection",
        ConfigValue(
            default=1 - 1e-10,
            domain=float,
            description="Default value for rejection of excluded species",
            doc="""Default value to be assigned as the rejection fraction for species
            identified as be excluded by membrane (high rejection).""",
        ),
    )

    def build(self):
        super().build()

        prop_params = self.config.property_package

        # Check that property package only supports single phase
        # TODO: Extend the model to handle phase equilibrium
        if len(prop_params.phase_list) > 1:
            raise ConfigurationError(
                "Nanofiltration0D model only supports single phase "
                "property packages."
            )

        # Check that the electroneutrality ion is a valid passing species
        # I.e. it must be in the passing species list or a monovalent ion
        if self.config.electroneutrality_ion is not None:
            try:
                bal_ion = prop_params.get_component(self.config.electroneutrality_ion)
            except AttributeError:
                raise ConfigurationError(
                    f"electroneutrality_ion ({self.config.electroneutrality_ion}) "
                    "is not a valid component in property package."
                )

            if self.config.passing_species_list is not None:
                if (
                    self.config.electroneutrality_ion
                    not in self.config.passing_species_list
                ):
                    raise ConfigurationError(
                        f"electroneutrality_ion ({self.config.electroneutrality_ion}) "
                        "must be a member of the passing species list."
                    )
            else:
                # Check that balancing ion is monovalent
                if (
                    not hasattr(bal_ion.config, "charge")
                    or abs(bal_ion.config.charge) > 1
                ):
                    raise ConfigurationError(
                        f"electroneutrality_ion ({self.config.electroneutrality_ion}) "
                        "must be a monovalent ion."
                    )

        # Check energy balance type is supported
        if self.config.energy_balance_type not in (
            EnergyBalanceType.none,
            EnergyBalanceType.isothermal,
        ):
            raise ConfigurationError(
                f"Nanofiltration0D model only supports isothermal operation or no energy balance "
                f"(assigned {self.config.energy_balance_type})"
            )

        # Check that pressure balance arguments are consistent
        if self.config.momentum_balance_type not in (
            MomentumBalanceType.none,
            MomentumBalanceType.pressureTotal,
        ):
            raise ConfigurationError(
                f"Nanofiltration0D model only supports total pressure balance or no pressure balance "
                f"(assigned {self.config.momentum_balance_type})"
            )
        elif (
            self.config.has_pressure_change
            and self.config.momentum_balance_type == MomentumBalanceType.none
        ):
            raise ConfigurationError(
                "Inconsistent configuration arguments. has_pressure_change=True "
                "requires that momentum_balance_type not equal MomentumBalanceType.none."
            )

        tmp_dict_in = dict(**self.config.property_package_args)
        tmp_dict_in["has_phase_equilibrium"] = False
        tmp_dict_in["defined_state"] = True

        self.properties_in = self.config.property_package.build_state_block(
            self.flowsheet().time, doc="Material properties at inlet", **tmp_dict_in
        )

        tmp_dict_out = dict(**self.config.property_package_args)
        tmp_dict_out["has_phase_equilibrium"] = False
        tmp_dict_out["defined_state"] = False

        self.properties_retentate = self.config.property_package.build_state_block(
            self.flowsheet().time, doc="Material properties at inlet", **tmp_dict_out
        )

        self.properties_permeate = self.config.property_package.build_state_block(
            self.flowsheet().time, doc="Material properties at inlet", **tmp_dict_out
        )

        self.add_port("inlet", self.properties_in, doc="Inlet Port")
        self.add_port("retentate", self.properties_retentate, doc="Retentate Port")
        self.add_port("permeate", self.properties_permeate, doc="Permeate Port")

        # NF separation variables
        self.recovery_solvent = Var(self.flowsheet().time, initialize=0.8)

        self._solute_set = Set(
            initialize=[
                j
                for j in prop_params.component_list
                if not prop_params.get_component(j).is_solvent()
                and j != self.config.electroneutrality_ion
            ]
        )

        def _init_rejection(b, _, j):
            if b.config.passing_species_list is not None:
                if j in b.config.passing_species_list:
                    return b.config.default_passing_rejection
                return b.config.default_excluded_rejection
            else:
                if j in prop_params.ion_set:
                    # Check charge
                    comp = prop_params.get_component(j)
                    if abs(comp.config.charge) == 1:
                        # Monovalent - include in set
                        return b.config.default_passing_rejection
                    else:
                        return b.config.default_excluded_rejection
                else:
                    # Non-ionic solute - assume passing species
                    return b.config.default_passing_rejection

        units = self.config.property_package.get_metadata().derived_units

        self.area = Var(
            initialize=1,
            bounds=(0.0, 1e6),
            units=units.LENGTH**2,
            doc="Membrane area",
        )
        self.rejection_comp = Var(
            self.flowsheet().time,
            self._solute_set,
            initialize=_init_rejection,
            bounds=(1e-10, 1),
            doc="Solute rejection fractions",
        )

        if self.config.has_pressure_change:
            self.deltaP = Var(
                self.flowsheet().time,
                initialize=0,
                units=units.PRESSURE,
                doc="Retentate side pressure change",
            )

        # Material balance
        @self.Constraint(self.flowsheet().time, prop_params.component_list)
        def material_balances(b, t, j):
            pset = b.properties_in[t].phase_list
            pcset = b.properties_in[t].phase_component_set
            return sum(
                b.properties_in[t].get_material_flow_terms(p, j)
                for p in pset
                if (p, j) in pcset
            ) == sum(
                b.properties_retentate[t].get_material_flow_terms(p, j)
                for p in pset
                if (p, j) in pcset
            ) + sum(
                b.properties_permeate[t].get_material_flow_terms(p, j)
                for p in pset
                if (p, j) in pcset
            )

        @self.Constraint(self.flowsheet().time, self.properties_in.phase_component_set)
        def separation_constraint(b, t, p, j):
            comp = prop_params.get_component(j)

            if comp.is_solvent():
                # Permeate flows equal to recovery * inlet flow
                return b.recovery_solvent[t] * b.properties_in[
                    t
                ].get_material_flow_terms(p, j) == b.properties_permeate[
                    t
                ].get_material_flow_terms(
                    p, j
                )
            elif j == self.config.electroneutrality_ion:
                # Rejection of electroneutrality ion will be solved using electroneutrality constraint
                return Constraint.Skip

            # else: Rejection = 1 - C_permeate/C_feed
            return (
                (1 - b.rejection_comp[t, j])
                * b.properties_in[t].conc_mol_phase_comp[p, j]
            ) == b.properties_permeate[t].conc_mol_phase_comp[p, j]

        if self.config.electroneutrality_ion is not None:
            # Add electroneutrality constraint for permeate stream
            @self.Constraint(self.flowsheet().time)
            def permeate_electronegativity(b, t):
                perm_state = b.properties_permeate[t]
                return 0 == sum(
                    perm_state.params.get_component(j).config.charge
                    * perm_state.flow_mol_phase_comp[p, j]
                    for p, j in perm_state.phase_component_set
                    if j in perm_state.params.ion_set
                )

        # Retentate pressure balance
        if self.config.momentum_balance_type == MomentumBalanceType.pressureTotal:

            @self.Constraint(self.flowsheet().time, doc="Retentate pressure balance")
            def retentate_pressure_balance(b, t):
                expr = b.properties_retentate[t].pressure
                if b.config.has_pressure_change:
                    expr += -b.deltaP[t]
                return b.properties_in[t].pressure == expr

        # Temperature equalities
        if self.config.energy_balance_type == EnergyBalanceType.isothermal:

            @self.Constraint(
                self.flowsheet().time, doc="Retentate temperature equality"
            )
            def retentate_temperature_equality(b, t):
                return (
                    b.properties_in[t].temperature
                    == b.properties_retentate[t].temperature
                )

            @self.Constraint(self.flowsheet().time, doc="Permeate temperature equality")
            def permeate_temperature_equality(b, t):
                return (
                    b.properties_in[t].temperature
                    == b.properties_permeate[t].temperature
                )

        @self.Expression(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.solvent_set,
            doc="Solvent mass transfer",
        )
        def flux_vol_solvent(b, t, p, j):
            if b.properties_in[t].get_material_flow_basis() == MaterialFlowBasis.mass:
                return -b.properties_permeate[t].flow_mass_phase_comp[p, j] / (
                    b.properties_in[t].dens_mass_phase[p] * b.area
                )
            elif (
                b.properties_in[t].get_material_flow_basis() == MaterialFlowBasis.molar
            ):
                # TODO- come up with better way to handle different locations of mw_comp in property models (generic vs simple ion prop model)
                if hasattr(b.properties_in[t].params, "mw_comp"):
                    mw_comp = b.properties_in[t].params.mw_comp[j]
                elif hasattr(b.properties_in[t], "mw_comp"):
                    mw_comp = b.properties_in[t].mw_comp[j]
                else:
                    raise ConfigurationError("mw_comp was not found.")
                return (
                    -b.properties_permeate[t].flow_mass_phase_comp[p, j] * mw_comp
                ) / (b.properties_in[t].dens_mass_phase[p] * b.area)

        @self.Expression(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.solvent_set | self._solute_set,
            doc="Mass-based component recovery",
        )
        def recovery_mass_phase_comp(b, t, p, j):
            return (
                b.properties_permeate[t].flow_mass_phase_comp[p, j]
                / b.properties_in[t].flow_mass_phase_comp[p, j]
            )

        @self.Expression(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            doc="Volumetric-based recovery",
        )
        def recovery_vol_phase(b, t, p):
            return (
                b.properties_permeate[t].flow_vol_phase[p]
                / b.properties_in[t].flow_vol_phase[p]
            )
